/**
 * Dev32lib.h
 *
 * Copyright (c) 2013-2015 David Azarewicz david@88watts.net
 *
 * The following source code is provided to you solely for the purpose of
 * assisting you in developing your own eCS device drivers. You may use
 * this software in your device drivers free of charge.
 */
#ifndef __DEV32LIB_H
#define __DEV32LIB_H

#define FAR16DATA __far16

#ifndef OS2_INCLUDED
#define INCL_NOPMAPI
#include <os2.h>
#endif

#pragma pack(1)

typedef struct {
  USHORT StratEntryOffset;
  USHORT EntryCS;
  USHORT IdcEntryOffset;
  USHORT IdcCS;
  USHORT Reserved;
  USHORT ProtDS;
} DDINFO;

typedef struct {
  ULONG ulNextDD;
  USHORT usAttribs;
  USHORT pfnStrategy;
  USHORT pfnIDC;
  char abName[8];
  USHORT usDevProtCS;     //Protected mode CS
  USHORT usDevProtDS;     //Protected mode DS
  USHORT usDevRealCS;     //Real mode CS (not used)
  USHORT usDevRealDS;     //Real mode DS (not used)
  ULONG ulCaps;
} DEV_HEADER;
#pragma pack()

#include <Dev32Help.h>
#include <strategy.h>
#include <Dev32rmbase.h>
#include <bsekee.h>

extern DEV_HEADER Header;

unsigned char PortInByte(unsigned short);
#pragma aux PortInByte = \
  "in   al,dx" \
  value [al] \
  parm nomemory [dx] \
  modify nomemory exact [al];

void PortOutByte(unsigned short, unsigned char);
#pragma aux PortOutByte = \
  "out  dx,al" \
  parm nomemory [dx] [al] \
  modify nomemory exact [];

unsigned short PortInWord(unsigned short);
#pragma aux PortInWord = \
  "in   ax,dx" \
  value [ax] \
  parm nomemory [dx] \
  modify nomemory exact [ax];

void PortOutWord(unsigned short, unsigned short);
#pragma aux PortOutWord = \
  "out  dx,ax" \
  parm nomemory [dx] [ax] \
  modify nomemory exact [];

unsigned long PortInDword(unsigned short);
#pragma aux PortInDword = \
  "in eax,dx" \
  parm nomemory [dx] value [eax] modify nomemory exact [eax];

void PortOutDword(unsigned short, unsigned long);
#pragma aux PortOutDword = \
  "out dx,eax" \
  parm nomemory [dx] [eax] modify nomemory exact [];

void DevCli(void);
#pragma aux DevCli = "cli";

void DevSti(void);
#pragma aux DevSti = "sti";

void DevClc(void);
#pragma aux DevClc = "clc";

void DevStc(void);
#pragma aux DevStc = "stc";

void DevInt3(void);
#pragma aux DevInt3 = "int 3";

unsigned long DevGetDS(void);
#pragma aux DevGetDS = "mov eax, ds";

unsigned long DevGetES(void);
#pragma aux DevGetES = "mov eax, es";

unsigned long DevGetSS(void);
#pragma aux DevGetSS = "mov eax, ss";

#pragma aux CastULONGToFar16 = "" parm nomemory [eax] value [eax] modify nomemory exact [];
void __far16 *CastULONGToFar16(ULONG);

#pragma aux CastFar16ToULONG = "" parm nomemory [eax] value [eax] modify nomemory exact [];
ULONG CastFar16ToULONG(void __far16 *);

/* PCI stuff ************************************************************/
#pragma pack(1)
#define MAXCAPCNT       32

typedef struct _PCI_CAP {
  UCHAR pos;
  UCHAR type;
} PCI_CAP, FAR * PPCI_CAP;

typedef struct _PCI_BAR {
  ULONG bar;
  ULONG start;
  ULONG size;
  unsigned io:1;
  unsigned type:2;
  unsigned prefetchable:1;
} PCI_BAR, FAR * PPCI_BAR;

struct PciDeviceInfo {
  /* Beginning of required section for Dev32lib */
  USHORT BusDevFunc;
  USHORT device;
  USHORT vendor;
  USHORT subsystem_vendor;
  USHORT subsystem_device;
  USHORT status;
  USHORT command;
  USHORT class_code_subclass;
  USHORT prog_if_rev_id;
  USHORT pcie_flags_reg; /* cached PCI-E Capabilities Register */
  UCHAR headertype;
  UCHAR irq;
  UCHAR ipin;
  UCHAR capcnt;
  UCHAR pcie_cap; /* PCI-E capability offset */
  UCHAR pcie_mpss; /* PCI-E Max Payload Size Supported */
  UCHAR pm_cap;
  UCHAR Unused;
  PCI_BAR bars[6];
  PCI_CAP caps[MAXCAPCNT];
  /* End of required section for Dev32lib */
};
typedef struct PciDeviceInfo PCI_DEVICEINFO;
#pragma pack()


/* Read a PCI config register
 * BusDevFunc: bits 15-8=Bus, bits 7-3=Device, bits 2-0=Function
 * reg: the register to read
 * size: the size of the read. 1=byte, 2=word, 4=dword
 * *data: a pointer to where to write the result
 * Returns: non-zero on failure.
 */
extern int __syscall PciReadConfig(const USHORT BusDevFunc, const UCHAR reg, const UCHAR size, void *data);

/* Write a PCI config register
 * BusDevFunc: bits 15-8=Bus, bits 7-3=Device, bits 2-0=Function
 * reg: the register to write
 * size: the size of the write. 1=byte, 2=word, 4=dword
 * data: the data to write
 * Returns: non-zero on failure.
 */
extern int __syscall PciWriteConfig(const USHORT BusDevFunc, const UCHAR reg, const UCHAR size, const ULONG data);

/* Find a PCI device by Vendor:Device ID
 * Vendor: the vendor ID to look for
 * Device: The device ID to look for
 * Index: Which device to return. zero based. 0 finds the first one, 1 finds the second one, etc.
 * Returns: 0xFFFF on failure. the BusDevFunc of the device if success.
 */
extern USHORT __syscall PciFindDevice(USHORT Vendor, USHORT Device, USHORT Index);

/* Find a PCI device by class ID
 * Class: the class ID to look for
 * Index: Which device to return. zero based. 0 finds the first one, 1 finds the second one, etc.
 * Returns: 0xFFFF on failure. the BusDevFunc of the device if success.
 */
extern USHORT __syscall PciFindClass(ULONG Class, USHORT Index);

/* Get the PCI device's configuration information
 * BusDevFunc: bits 15-8=Bus, bits 7-3=Device, bits 2-0=Function
 * pDevInfo: Pointer to the PCI_DEVICEINFO structure where the information is returned.
 * Returns: non-zero on failure.
 */
extern int __syscall PciGetDeviceInfo(USHORT BusDevFunc, PCI_DEVICEINFO *pDevInfo);
extern int __syscall PciSetPowerStateD0(PCI_DEVICEINFO *pDevInfo);
extern int __syscall PciSetPowerStateD3hot(PCI_DEVICEINFO *pDevInfo);
extern int __syscall PciSetBusMaster(PCI_DEVICEINFO *pDevInfo);
extern UCHAR __syscall PciGetCapPos(const UCHAR type, PCI_DEVICEINFO *pDevInfo);

/* Resource Manager stuff ***********************************************/

extern int __syscall RmAddMem(HDRIVER hDriver, HRESOURCE *phResource, ULONG addr, ULONG length);
extern int __syscall RmAddIrq(HDRIVER hDriver, HRESOURCE *phResource, USHORT irq, USHORT pin);
extern int __syscall RmAddIo(HDRIVER hDriver, HRESOURCE *phResource, USHORT addr, USHORT length);

/* The Rm1 functions are simplified for drivers that only support one adapter.
 * This avoids the necessity for the driver to manage all the handles.
 */

/* Creates a driver. Must be called first */
extern int __syscall Rm1CreateDriver(DRIVERSTRUCT *pDrs);

/* Destroys everything that was created */
extern void __syscall Rm1Destroy(int iDestroyDriver);

/* Add the specified type of resource */
extern int __syscall Rm1AddMem(ULONG addr, ULONG length);
extern int __syscall Rm1AddIrq(USHORT irq, USHORT pin);
extern int __syscall Rm1AddIo(USHORT addr, USHORT length);

/* Creates the adapter. Must be called after adding the resources. */
extern int __syscall Rm1CreateAdapter(ADAPTERSTRUCT *pAds);

/* Memory functions *****************************************************/
#define PAGE_SIZE 0x1000

/* Maps a physical address to a linear address. This function is for mapping
 * the physical register space of a device to a linear address so it can be
 * accessed by a device driver. Mapping is always done in whole pages so if
 * you map 16 bytes, you will really get 4096 bytes mapped. The conversion
 * functions in this library cannot convert pointers returned by this
 * function to Far16. This function cannot be called at interrupt time.
 */
extern void * __syscall MapPhysToLin(ULONG ulPhysAdr, ULONG ulSize);

/* This function cannot be called at interrupt time */
extern void * __syscall D32VMAlloc(ULONG ulSize);

/* This function cannot be called at interrupt time */
extern int __syscall D32VMFree(void *LinAdr);

/* This function cannot be called at interrupt time */
extern void * __syscall D32AllocMem(ULONG ulSize, ULONG *pPhysAdr);

/* This function cannot be called at interrupt time */
extern int __syscall D32FreeMem(void *LinAdr);

/* The Mem* functions are special memory allocation functions that maintain
 * a pool of memory in such a way that they can provide linear, physical
 * and far16 pointers to every buffer that is allocated. The maximum size
 * of memory that can be allocated is 64K.
 */

/* Allocate some memory and return a flat linear address to it. Max 64K size
 * This function cannot be called at interrupt time.
 *
 * The amount of memory returned is the size of memory requested rounded up to
 * the nearest paragraph (16 bytes). If you request 15 bytes, you get 16 bytes.
 * If you request 48 bytes, you get 48 bytes.
 *
 * All memory returned by MemAlloc is guaranteed to be paragraph (16 byte) aligned.
 */
void * __syscall MemAlloc(ULONG ulSize);

/* Free the memory that was allocated using MemAlloc()
 * This function can be called at interrupt time.
 */
int __syscall MemFree(void *pLinAdr);

/* Return the physical address of memory that was allocated using MemAlloc()
 * This function can be called at interrupt time.
 */
ULONG __syscall MemPhysAdr(void *pLinAdr);

/* Return the far16 pointer to the memory that was allocated using MemAlloc()
 * This function can be called at interrupt time.
 */
void __far16 * __syscall MemFar16Adr(void *pLinAdr);

/* some Debug stuff *****************************************************/

/* The debug functions output to the serial port */
extern int D32g_DbgLevel;
extern int D32g_DbgBufWrap;
extern USHORT D32g_ComBase;
extern void __syscall dprintf(int level, const char *fmt, ...);
extern void __syscall dHexDump(int level, const void *p, int len, const char *fmt, ...);
extern unsigned long __syscall dCopyToUser(unsigned char *Buffer, unsigned long Count);
extern void __syscall InitComPort(ULONG BaudRate);

/* Trace functions ******************************************************/

extern void __syscall TraceInit(ULONG ulMajor, ULONG ulLevel);

/* Warning: This function can use a lot of stack space */
extern void __syscall TraceBuf(ULONG minCode, ULONG DataLen, void *pBuf);

extern void __syscall TraceArgs(ULONG minCode, ULONG DataLen, ...);

/* Misc functions *******************************************************/

typedef struct {
  volatile ULONG *pMsCount;
  ULONG Start;
  ULONG End;
} TIMER;

extern ULONG __syscall TimeGetSec(void);
extern ULONG __syscall TimeGetMs(void);
extern void __syscall TimeReset(void);
extern void __syscall TimeInit(void);
extern ULONG __syscall TimeDiff(void);
extern void __syscall msleep(ULONG msecs);
extern void TimerInit(TIMER *pTimer, ULONG Milliseconds);
extern int TimerCheckAndBlock(TIMER *pTimer);

/* Copies the specified string to the device driver header, enforcing the
 * rules (8 char max, blank padded)
 */
extern void __syscall UtSetDriverName(char *pDevName);

/* Modifies the given string assuming it is a device driver name.
 * Finds the first '$'.
 * Checks the character immediately before the '$' to see if it is numeric.
 * If it is numeric, it is replaced with '0'+iChar. If not, '0'+iChar is
 * inserted immediately before the '$'. Non-zero status is returned on any
 * failure: iChar is not between 0 and 9, inserting a char would make the
 * name longer than 8 characters, the name is only 1 character. The buffer
 * must be at least 9 bytes long to hold the terminating null character.
 */
extern int __syscall UtModifyName(char *pDevName, int iChar, int iOverwrite);

/* Compare the given command line pointer to the given argument string and
 * return true if there is a match. ArgString should be all upper case.
 */
extern int __syscall ArgCmp(const char *CmdString, const char *ArgString);

/* Convert a string to a long value using the specified base.
 * pString is the string to be converted.
 * pEnd is a pointer to the string where conversion ended. Can be NULL.
 * Base is the base of the conversion. If 0 the function determines the base
        automatically. octal(begins with a 0) or hex(begins with 0x), or decimal.
 */
extern long __syscall strtol(const char *pString, const char **pEnd, int Base);

extern int __syscall GetString(const CHAR *pStringIn, CHAR *pTarget, ULONG MaxLen);

/* This routine is obsolete and will eventually be removed */
extern ULONG __syscall GetHex(CHAR *pString, UCHAR *pResult, ULONG MaxLen);

static inline ULONG ReadDword(void *address) { return *(ULONG*)address; }
static inline USHORT ReadWord(void *address) { return *(USHORT*)address; }
static inline UCHAR ReadByte(void *address) { return *(UCHAR*)address; }
static inline void WriteDword(void *address, ULONG value) { *(ULONG*)address = value; }
static inline void WriteWord(void *address, USHORT value) { *(USHORT*)address = value; }
static inline void WriteByte(void *address, UCHAR value) { *(UCHAR*)address = value; }


#pragma aux D32ThunkStackTo32 modify []
extern void D32ThunkStackTo32(void);

/* Converts flat pointer to a Far16 pointer. This function can covert data segment pointers
 * and stack segment pointers. Attempting to convert a flat pointer that cannot be
 * converted will generate a trap 3. Use of D32ThunkStackTo32() is required instead of
 * KernThunkStackTo32() in order for FlatToFar16() to work properly for stack variables.
 */
#pragma aux FlatToFar16 parm [eax] modify exact [eax]
extern void __far16 *FlatToFar16(void *Pointer);

/* Converts any Far16 pointer to a flat pointer */
#pragma aux Far16ToFlat parm [eax] modify exact [eax]
extern void *Far16ToFlat(void __far16 *Pointer);

#pragma aux TestAndSetBit parm [eax] [ebx] value [eax] ;
extern ULONG TestAndSetBit(ULONG BitNumbar, volatile ULONG *addr);

#pragma aux ClearBit parm [eax] [ebx] modify [eax] ;
extern void ClearBit(ULONG BitNumber, volatile ULONG *addr);

#pragma aux LockInc parm [eax] modify [eax] ;
extern void LockInc(volatile ULONG *addr);

#pragma aux LockDecAndTest parm [eax] value [eax] modify [eax] ;
extern int LockDecAndTest(volatile ULONG *addr);

/* These are spin loop functions. Avoid using these if possible.
 * Use ProcBlock instead.
 */
#pragma aux udelay parm [ecx] modify [eax ecx] ;
extern void udelay(ULONG microseconds);
extern void __syscall mdelay(ULONG milliseconds);

/* The following File I/O functions are experimental and may not work, may be changed,
 * or may be removed.
 */
extern int __syscall QueueFileOpen(char *pFileName);
extern int __syscall ReadQueuedFile(char *pFileName, void *pBuffer, ULONG ulBufSize, ULONG *pulActual);

/* This function is only available in debug builds */
extern void __syscall MemHeapVerify(void *LinAdr);

/* some libc stuff ******************************************************/

/* ctype macros */
#define isupper(ch) ((ch) >= 'A' && (ch) <= 'Z')
#define islower(ch) ((ch) >= 'a' && (ch) <= 'z')
extern int __syscall tolower(int ch);
extern int __syscall toupper(int ch);

#ifndef __size_t
#define __size_t
typedef unsigned long size_t;
#endif

void *memchr(const void *s, short c, size_t n);
int memcmp(const void *s1, const void *s2, size_t n);
void *memcpy(void *s1, const void *s2, size_t n);
void *memset(void *s, short c, size_t n);
char *strcat(char *s1, const char *s2);
int strcmp(const char *s1, const char *s2);
char *strcpy(char *s1, const char *s2);
size_t strlcpy(char *dst, const char *src, size_t siz);
int strlen(const char *s);
#pragma intrinsic(memchr,memcmp,memcpy,memset)
#pragma intrinsic(strcat,strcmp,strlen,strcpy)

extern int __syscall stricmp(char *str0, char *str1);
extern int __syscall atoi(char *str);

#ifndef min
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

/*
 * Very basic formatted string functions
 * leading zero modifer is supported.
 * single digit length modifier is supported.
 * modifiers are: l=long F=far
 * functions are: s=string, c=char, d=decimal, x=hex, p=pointer
 * Examples:
 * %s = near (flat) string
 * %Fs = far16 string
 * %d = 32 bit decimal
 * %ld = 32 bit decimal
 * %x = 32 bit hex
 * %lx = 32 bit hex
 * %p = 32 bit flat pointer
 * %Fp or %lp = 16:16 pointer
 */
extern int __syscall sprintf(char *buf, const char *fmt, ...);
extern int __syscall snprintf(char *buf, int len, const char *fmt, ...);

/* DisplayMessage uses the DevHelp_Save_Message call to output a message to
 * the console. This can be used by DEVICE and BASEDEV drivers.
 */
extern int DisplayMessage(char *pMessage);
#pragma aux DisplayMessage parm [eax] value [eax] modify nomemory [edx];

/* iprintf formats a string and calls DisplayMessage to output to the console.
 */
extern void __syscall iprintf(const char *fmt, ...);
#endif //__DEV32LIB_H
